import argparse
import asyncio
import logging
import os

from .executors import EXECUTORS, load_executor
from .streamers import STREAMERS, load_streamer
from .core import streamline

logger = logging.getLogger(__file__)

def streamline_command():
    cmd_parser = argparse.ArgumentParser(prog='streamline')
    cmd_parser.add_argument('--input', help='Set source (Default stdin)', default='-')
    cmd_parser.add_argument('--output', help='Set target of output (Default stdout)', default='-')
    cmd_parser.add_argument(
        '-n', '--headers-only',
        help='Only show input values with truthy non-exception results (useful with --filter)',
        action='store_true',
        default=False,
    )
    cmd_parser.add_argument(
        '-m', '--executor',
        help='Execution Module',
        default=None,
    )
    cmd_parser.add_argument(
        '-s', '--streamer',
        help='Execution Module',
        default='line',
    )
    cmd_parser.add_argument(
        '-w', '--workers',
        type=int,
        help='Number of concurrent workers for execution modules',
        default=10,
    )
    cmd_parser.add_argument(
        '-d', '--headers',
        action='store_true',
        help='Prefix output with entry name',
        default=False,
    )
    cmd_parser.add_argument(
        '-p', '--progress',
        action='store_true',
        help='Output progress bar and dont display results until the end',
        default=False,
    )
    cmd_parser.add_argument(
        '-e', '--extract',
        help='Extract a part of the result by dot path (e.g. result.foobar)',
        default=None,
    )
    cmd_parser.add_argument(
        '--stacktraces',
        action='store_true',
        default=False,
        help='Extract a part of the result by dot path (e.g. result.foobar)',
    )
    args, extra_args = cmd_parser.parse_known_args()

    Streamer = load_streamer(args.streamer)
    executor = load_executor(args.executor, extra_args)
    streamer = Streamer(args.input, args.output)
    loop = asyncio.get_event_loop()
    task = asyncio.ensure_future(streamline(
        streamer,
        executor=executor,
        headers=args.headers,
        progress=args.progress,
        extract=args.extract,
        stacktraces=args.stacktraces,
        worker_count=args.workers,
    ))
    loop.run_until_complete(task)



import json
import sys

from .utils import import_obj

def _get_file_io(name, write=False):
    if name == '-' and write:
        return sys.stdout
    elif name == '-' and not write:
        return sys.stdin
    else:
        return open(name, 'w' if write else 'r', 1)


class UninitializedStreamerError(Exception):
     default_message = 'This is a default message!'

     def __init__(self, message=None, **kwargs):
         super().__init__(message or default_message, **kwargs)


class LineStreamer():
    DELIMITER = '\n'

    def __init__(self, source_name, target_name, force_flush=False):
        self.source_name = source_name
        self.target_name = target_name
        self.force_flush = force_flush

        self.first_written = False
        self.source = None
        self.target = None
        self.target_template = None

    def __iter__(self):
        return self.read()

    def __enter__(self):
        self.source = _get_file_io(self.source_name)
        if '{entry}' in self.target_name:
            self.target_template = self.target_name
        else:
            self.target = _get_file_io(self.target_name, write=True)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if hasattr(self.source, 'close'):
            self.source.close()
        if self.target and hasattr(self.target, 'close'):
            self.target.close()

    def read(self):
        if self.source is None:
            raise UninitializedStreamerError()

        for line in self.source:
            yield line.rstrip('\n')

    def write(self, entry, result):
        if not isinstance(result, str):
            result = json.dumps(result)

        if self.first_written:
            result = self.DELIMITER + result
        else:
            self.first_written = True

        if self.target_template:
            with open(self.target_template.format(entry=entry), 'w') as target_file:
                target_file.write(result)
        elif self.target is None:
            raise UninitializedStreamerError()
        else:
            self.target.write(result)
            if self.force_flush and hasattr(self.target, 'flush'):
                self.target.flush()
        
STREAMERS = {
    'line': LineStreamer,
}

def load_streamer(path):
    if path is None:
        return None
    elif '.' in path:
        Streamer = import_obj(path)
    else:
        Streamer = STREAMERS.get(path)
    return Streamer



from importlib import import_module
import sys
import os

def import_obj(path):
    current_dir = os.path.abspath('.')
    if current_dir not in sys.path:
        sys.path.append(current_dir)
    module_path, handler_name = path.rsplit('.', 1)
    module = import_module(module_path)
    for attr in module_path.split('.')[1:]:
        module = getattr(module, attr)
    handler_obj = getattr(module, handler_name)
    return handler_obj

def inject_module(module_name, namespace):
    try:
        module = import_module(module_name)
    except ModuleNotFoundError as e:
        print('This plugin requires module: "{}"\nTry "pip install {}"'.format(module_name, module_name))
        sys.exit(6)
    sys.modules[module_name] = module
    namespace[module_name] = module




from importlib import import_module
import asyncio
import logging
import traceback
import json
import sys

from .extractor import extract_path, parse_selectors

logger = logging.getLogger(__file__)

class AsyncExecutor():
    """
        :: Worker-oriented event loop processor

        Workers are no longer a necessary concept in an asynchronous world. However, the concept can still be very
        helpful for controlling resource usage on the host machine or remote systems used by the job. For this reason
        I'm re-implementing a worker-style executor pool which could be used with jobs that are threaded or async.
    """
    DEFAULT_WORKERS = 5

    def __init__(self, source, target, executor, show_progress=True, stacktraces=False, extract=None, workers=DEFAULT_WORKERS):
        self.source = source
        self.target = target
        self.extract = extract
        self.executor = executor
        self.input_queue = asyncio.Queue()
        self.output_queue = asyncio.Queue()
        self.show_progress = show_progress
        self.worker_count = workers or self.DEFAULT_WORKERS
        self.stacktraces = stacktraces
        self.loop = asyncio.get_event_loop()

        # State data
        self.entry_count = 0
        self.complete_count = 0
        self.active_count = 0
        self.all_enqueued = False

    def _show_progress(self, newline=False):
         if not self.show_progress:
             return

         sys.stdout.write('\r {done} out of {total} tasks complete - {percent_done}% (running={running}; pending={pending})'.format(
             done=self.complete_count,
             total=self.entry_count,
             percent_done=int(self.complete_count / self.entry_count * 100),
             running=self.active_count,
             pending=self.input_queue.qsize(),
         ))
         if newline:
             # Finish with a newline
             print('')

    def _save_result(self, entry, result):
        self.complete_count += 1
        self.output_queue.put_nowait((entry, result))

    def start(self):
        # Enqueue all work tasks
        for entry in self.source:
            self.entry_count += 1
            self.input_queue.put_nowait(entry.rstrip('\n'))
        self.all_enqueued = True

        # Start workers
        workers = []
        for i in range(self.worker_count):
            worker = self.loop.create_task(self._worker())
            workers.append(worker)
            asyncio.ensure_future(worker)

        #self.loop.run_until_complete(self._execute_jobs())

    def _is_complete(self):
        return self.complete_count == self.entry_count

    async def _execute_jobs(self):
        # Wait and show progress
        while True:
            self._show_progress()
            await asyncio.wait(workers, timeout=0.1)
            if self._is_complete():
                break
            else:
                self._show_progress()
        self._show_progress(newline=True)

    async def stream_results(self):
        while not self._is_complete() or self.output_queue.qsize():
            try:
                result_pair = await asyncio.wait_for(self.output_queue.get(), timeout=.1)
            except asyncio.TimeoutError:
                continue
            yield result_pair
            self._show_progress()
        self._show_progress(newline=True)

    async def _worker(self):
        while True:
            try:
                entry = await asyncio.wait_for(self.input_queue.get(), timeout=.1)
            except asyncio.TimeoutError:
                if self.input_queue.qsize() == 0 and self.all_enqueued:
                    return
                else:
                    continue
            self.active_count += 1
            try:
                if asyncio.iscoroutinefunction(self.executor):
                    result = await self.executor(entry)
                else:
                    def executor_wrapper():
                        return self.executor(entry)
                    result = await self.loop.run_in_executor(None, executor_wrapper)
            except Exception as e:
                if self.stacktraces:
                    result = '{}'.format(traceback.format_exc())
                else:
                    result = str(e)
            self._save_result(entry, result)
            self.input_queue.task_done()
            self.active_count -= 1

async def noop_executor(source):
    for entry in source:
        yield entry, entry

async def streamline(Streamer, executor=None, headers=None, progress=False, extract=None, worker_count=None, stacktraces=True):
    if extract is None:
        extractor = lambda x: x
    else:
        selectors = parse_selectors(extract)
        extractor = lambda x: extract_path(x, selectors)

    with Streamer as streamer:
        if executor:
            ae = AsyncExecutor(
                streamer.read(),
                streamer.write,
                executor,
                show_progress=progress,
                stacktraces=stacktraces,
                workers=worker_count,
            )
            ae.start()
            result_stream = ae.stream_results()
        else:
            result_stream = noop_executor(streamer.read())
            
        async for entry, result in result_stream:
            extracted_result = extractor(result)

            if headers:
                if not isinstance(extracted_result, str):
                    extracted_result = json.dumps(entry)
                extracted_result = '{}: {}'.format(entry, extracted_result)
            streamer.write(entry, extracted_result)
